///////////////////////////////////////////////////////////////////////////
//
// NAME
//  Project1.cpp -- command-line (shell) interface to project 1 code
//
// SYNOPSIS
//  Project1 cylWarp input.tga output.tga f [k1 k2]
//  Project1 LucasKanade input1.tga input2.tga tx ty nLevels nSteps
//  Project1 blendPairs pairlist.txt outfile.tga blendWidth
//  Project1 script script.cmd
//
// PARAMTERS
//  input*.tga      input image
//  output.tga      output image
//
//  f               focal length in pixels
//  k1, k2          radial distortion parameters
//
//  tx, ty          initial translation guess
//  nLevels         number of levels in hierarchical alignment algorithm
//  nSteps          number of iterative correction steps per level
//
//  pairlist.txt    file of image pair names and relative translations
//                  this is usually the concatenation of outputs from LucasKanade
//
//  blendWidth		width of the horizontal blending function
//
//  script.cmd      script file (command line file)
//
// DESCRIPTION
//  This file contains a set of simple command line processes from which
//  you can build a simple automatic cylindrical stitching application.
//
//  Use the programs here to perform a series of operations such as:
//  1. warp all of the images into cylindrical coordinate (and undo radial distortion)
//  2. align pairs of images using a pyramid motion estimator
//  3. read in all of the images and perform pairwise blends
//     to obtain a final (rectified and trimmed) mosaic
//
// TIPS
//  To become familiar with this code, single-step through a couple of
//  examples.  Also, if you are running inside the debugger, place
//  a breakpoint inside the catch statement and at the last statement
//  in main() so you can see the error message before the program exits.
//
// SEE ALSO
//  Image.h             basic image class documentation
//
// (modified for CSE455 Winter 2003)
///////////////////////////////////////////////////////////////////////////

#include "../ImageLib/ImageLib.h"
#include "WarpCylindrical.h"
#include "LucasKanade.h"
#include "BlendImages.h"

int main(int argc, const char *argv[]);     // forward declaration

int CylWarp(int argc, const char *argv[])
{
    // Warp the input image to correct for radial distortion and/or map to cylindrical coordinates
    if (argc < 5)
    {
        printf("usage: %s input.tga output.tga f [k1 k2]\n", argv[1]);
        return -1;
    }
    const char *infile  = argv[2];
    const char *outfile = argv[3];
    float f             = (float) atof(argv[4]);
    float k1            = (argc > 5) ? (float) atof(argv[5]) : 0.0f;
    float k2            = (argc > 6) ? (float) atof(argv[6]) : 0.0f;

    CByteImage src, dst;
    ReadFile(src, infile);
    if (src.Shape().nBands != 4)
        src = ConvertToRGBA(src);
    CShape sh = src.Shape();
    CFloatImage uv = WarpCylindricalField(sh, sh, f, k1, k2);
    WarpLocal(src, dst, uv, false, eWarpInterpLinear);
    WriteFile(dst, outfile);
    return 0;
}

int LucasKanade(int argc, const char *argv[])
{
    // Align two images using a hierarchical (coarse-to-fine) Lucas-Kanade with search
    if (argc < 8)
    {
        printf("usage: %s input1.tga input2.tga tx ty nLevels nSteps\n", argv[1]);
        return -1;
    }
    const char *infile1 = argv[2];
    const char *infile2 = argv[3];
    int tx              = atoi(argv[4]);
    int ty              = atoi(argv[5]);
    int nLevels         = atoi(argv[6]);
    int nSteps          = atoi(argv[7]);

    CByteImage img1, img2;
    ReadFile(img1, infile1);
    ReadFile(img2, infile2);
    float t[2] = {(float) tx, (float) ty};
    PyramidalLucasKanade(img1, img2, t, nLevels, nSteps);

    // Print out the result
    printf("%s %s %.2f %.2f\n", infile1, infile2, t[0], t[1]);

    return 0;
}

int BlendPairs(int argc, const char *argv[])
{
    // Align two images using a hierarchical (coarse-to-fine) Lucas-Kanade with search
    if (argc < 5)
    {
        printf("usage: %s pairlist.txt outimg.tga blendWidth", argv[1]);
        return -1;
    }
    const char *pairlist= argv[2];
    const char *outfile = argv[3];
    float blendWidth    = (float) atof(argv[4]);

    // Open the list of image pairs
    FILE *stream = fopen(pairlist, "r");
    if (stream == 0)
        throw CError("%s: could not open the file %s", argv[1], pairlist);

    // Construct the list of images and translations
    CImagePositionV ipList;
    char line[1024], infile1[1024], infile2[1024];
    float rel_t[2];
    int n;
    for (n = 0; fgets(line, 1024, stream); n++)
    {
        // Compute the position from the PREVIOUS displacement
        CImagePosition ip;
        for (int k = 0; k < 2; k++)
            ip.position[k] = (n > 0) ? ipList[n-1].position[k] - rel_t[k] : 0.0f;

        // Read the line and push the image onto the list
        if (sscanf(line, "%s %s %f %f", infile1, infile2,
                   &rel_t[0], &rel_t[1]) != 4)
            throw CError("%s: error reading %s", argv[1], pairlist);
        ReadFile(ip.img, infile1);
        ipList.push_back(ip);
    }

    // Read in the last image
    CImagePosition ip;
    for (int k = 0; k < 2; k++)
        ip.position[k] = ipList[n-1].position[k] - rel_t[k];
    ReadFile(ip.img, infile2);
    ipList.push_back(ip);
    fclose(stream);

    CByteImage result = BlendImages(ipList, blendWidth);
    WriteFile(result, outfile);
    return 0;
}

int Script(int argc, const char *argv[])
{
    // Read a series of commands from a script file
    //  (an alternative to a shell-level command file)
    if (argc < 3)
    {
        printf("usage: %s script.cmd\n", argv[1]);
        return -1;
    }
    FILE *stream = fopen(argv[2], "r");
    if (stream == 0)
        throw CError("Could not open %s", argv[2]);

    // Process each command line
    char line[1024];
    while (fgets(line, 1024, stream))
    {
        fprintf(stderr, line);
        if (line[0] == '/' && line[1] == '/')
            continue;   // skip the comment line
        char *ptr = line;
        char *argv2[256];
        int argc2;
        for (argc2 = 0; argc2 < 256 && *ptr; argc2++)
        {
            while (*ptr && isspace(*ptr)) ptr++;
            argv2[argc2] = ptr;
            while (*ptr && !isspace(*ptr)) ptr++;
            if (*ptr)
                *ptr++ = 0;     // null terminate the argument
        }
        if (argc2 < 2)
            continue;

        // Call the dispatch routine
        int code = main(argc2, (const char **) argv2);
        if (code)
            return code;
    }
    fclose(stream);
    return 0;
}

int main(int argc, const char *argv[])
{
    try
    {
        // Branch to processing code based on first argument
        if (argc > 1 && strcmp(argv[1], "cylWarp") == 0)
            return CylWarp(argc, argv);
        else if (argc > 1 && strcmp(argv[1], "LucasKanade") == 0)
            return LucasKanade(argc, argv);
        else if (argc > 1 && strcmp(argv[1], "blendPairs") == 0)
            return BlendPairs(argc, argv);
        else if (argc > 1 && strcmp(argv[1], "script") == 0)
            return Script(argc, argv);
		else {
			printf("usage: \n");
	        printf("	v4gP1 cylWarp input.tga output.tga f [k1 k2]\n");
			printf("	v4gP1 LucasKanade input1.tga input2.tga tx ty nLevels nSteps\n");
			printf("	v4gP1 blendPairs pairlist.txt outimg.tga blendWidth\n");
			printf("	v4gP1 script script.cmd\n");
		}
    }
    catch (CError &err) {
        printf(err.message);
        return -1;
    }
    return 0;
}
